@@ -5,8 +5,9 @@ use std::io;
 use std::result;
 use std::fmt;
 
-pub struct Pkcs12(openssl::pkcs12::ParsedPkcs12);
-pub struct Certificate(openssl::x509::X509);
+use tls_api::Error;
+use tls_api::Result;
+
 
 pub struct TlsConnectorBuilder(openssl::ssl::SslConnectorBuilder);
 pub struct TlsConnector(openssl::ssl::SslConnector);
@@ -14,40 +15,48 @@ pub struct TlsConnector(openssl::ssl::SslConnector);
 pub struct TlsAcceptorBuilder(openssl::ssl::SslAcceptorBuilder);
 pub struct TlsAcceptor(openssl::ssl::SslAcceptor);
 
-use tls_api::Error;
-use tls_api::Result;
 
+// TODO: https://github.com/sfackler/rust-openssl/pull/646
+#[cfg(has_alpn)]
+pub const HAS_ALPN: bool = true;
+#[cfg(not(has_alpn))]
+pub const HAS_ALPN: bool = false;
 
-fn map_error_stack(e: openssl::error::ErrorStack) -> Error {
-    Error::new(e)
-}
-
-
-impl tls_api::Pkcs12 for Pkcs12 {
-    fn from_der(der: &[u8], password: &str) -> Result<Self> {
-        let pkcs12 = openssl::pkcs12::Pkcs12::from_der(der).map_err(map_error_stack)?;
-        let parsed = pkcs12.parse(password).map_err(map_error_stack)?;
-        Ok(Pkcs12(parsed))
-    }
-}
-
-impl tls_api::Certificate for Certificate {
-    fn from_der(der: &[u8]) -> Result<Self> where Self: Sized {
-        openssl::x509::X509::from_der(der)
-            .map(Certificate)
-            .map_err(Error::new)
-    }
-}
 
 impl tls_api::TlsConnectorBuilder for TlsConnectorBuilder {
     type Connector = TlsConnector;
 
-    fn add_root_certificate(&mut self, cert: Certificate) -> Result<&mut Self> {
+    type Underlying = openssl::ssl::SslConnectorBuilder;
+
+    fn underlying_mut(&mut self) -> &mut openssl::ssl::SslConnectorBuilder {
+        &mut self.0
+    }
+
+    fn supports_alpn() -> bool {
+        HAS_ALPN
+    }
+
+    #[cfg(has_alpn)]
+    fn set_alpn_protocols(&mut self, protocols: &[&[u8]]) -> Result<()> {
+        self.0.builder_mut().set_alpn_protocols(protocols)
+            .map_err(Error::new)
+    }
+
+    #[cfg(not(has_alpn))]
+    fn set_alpn_protocols(&mut self, protocols: &[&[u8]]) -> Result<()> {
+        Err(Error::new_other("openssl is compiled without alpn"))
+    }
+
+    fn add_root_certificate(&mut self, cert: tls_api::Certificate) -> Result<&mut Self> {
+        let cert = openssl::x509::X509::from_der(&cert.into_der())
+            .map_err(Error::new)?;
+
         self.0
             .builder_mut()
             .cert_store_mut()
-            .add_cert(cert.0)
-                .map_err(map_error_stack)?;
+            .add_cert(cert)
+            .map_err(Error::new)?;
+        
         Ok(self)
     }
 
@@ -56,6 +65,12 @@ impl tls_api::TlsConnectorBuilder for TlsConnectorBuilder {
     }
 }
 
+impl TlsConnectorBuilder {
+    pub fn builder_mut(&mut self) -> &mut openssl::ssl::SslConnectorBuilder {
+        &mut self.0
+    }
+}
+
 #[derive(Debug)]
 struct TlsStream<S : io::Read + io::Write + fmt::Debug>(openssl::ssl::SslStream<S>);
 
@@ -94,6 +109,16 @@ impl<S : io::Read + io::Write + fmt::Debug + Send + Sync + 'static> tls_api::Tls
     fn get_mut(&mut self) -> &mut S {
         self.0.get_mut()
     }
+
+    #[cfg(has_alpn)]
+    fn get_alpn_protocol(&self) -> Option<Vec<u8>> {
+        self.0.ssl().selected_alpn_protocol().map(Vec::from)
+    }
+
+    #[cfg(not(has_alpn))]
+    fn get_alpn_protocol(&self) -> Option<Vec<u8>> {
+        None
+    }
 }
 
 
@@ -135,7 +160,6 @@ fn map_handshake_error<S>(e: openssl::ssl::HandshakeError<S>) -> tls_api::Handsh
 
 impl tls_api::TlsConnector for TlsConnector {
     type Builder = TlsConnectorBuilder;
-    type Certificate = Certificate;
 
     fn builder() -> Result<TlsConnectorBuilder> {
         openssl::ssl::SslConnectorBuilder::new(openssl::ssl::SslMethod::tls())
@@ -164,27 +188,62 @@ impl tls_api::TlsConnector for TlsConnector {
     }
 }
 
+
+// TlsAcceptor and TlsAcceptorBuilder
+
+impl TlsAcceptorBuilder {
+    pub fn from_pkcs12(pkcs12: &[u8], password: &str) -> Result<TlsAcceptorBuilder> {
+        let pkcs12 = openssl::pkcs12::Pkcs12::from_der(pkcs12).map_err(Error::new)?;
+        let pkcs12 = pkcs12.parse(password).map_err(Error::new)?;
+
+        openssl::ssl::SslAcceptorBuilder::mozilla_intermediate(
+            openssl::ssl::SslMethod::tls(),
+            &pkcs12.pkey,
+            &pkcs12.cert,
+            &pkcs12.chain)
+                .map(TlsAcceptorBuilder)
+                .map_err(Error::new)
+    }
+}
+
 impl tls_api::TlsAcceptorBuilder for TlsAcceptorBuilder {
     type Acceptor = TlsAcceptor;
 
+    type Underlying = openssl::ssl::SslAcceptorBuilder;
+
+    fn underlying_mut(&mut self) -> &mut openssl::ssl::SslAcceptorBuilder {
+        &mut self.0
+    }
+
+    fn supports_alpn() -> bool {
+        HAS_ALPN
+    }
+
+    #[cfg(has_alpn)]
+    fn set_alpn_protocols(&mut self, protocols: &[&[u8]]) -> Result<()> {
+        self.0.builder_mut().set_alpn_protocols(protocols)
+            .map_err(Error::new)
+    }
+
+    #[cfg(not(has_alpn))]
+    fn set_alpn_protocols(&mut self, protocols: &[&[u8]]) -> Result<()> {
+        Err(Error::new_other("openssl is compiled without alpn"))
+    }
+
+
     fn build(self) -> Result<TlsAcceptor> {
         Ok(TlsAcceptor(self.0.build()))
     }
 }
 
-impl tls_api::TlsAcceptor for TlsAcceptor {
-    type Pkcs12 = Pkcs12;
-    type Builder = TlsAcceptorBuilder;
-
-    fn builder(pkcs12: Pkcs12) -> Result<TlsAcceptorBuilder> {
-        openssl::ssl::SslAcceptorBuilder::mozilla_intermediate(
-            openssl::ssl::SslMethod::tls(),
-            &pkcs12.0.pkey,
-            &pkcs12.0.cert,
-            &pkcs12.0.chain)
-                .map(TlsAcceptorBuilder)
-                .map_err(map_error_stack)
+impl TlsAcceptorBuilder {
+    pub fn builder_mut(&mut self) -> &mut openssl::ssl::SslAcceptorBuilder {
+        &mut self.0
     }
+}
+
+impl tls_api::TlsAcceptor for TlsAcceptor {
+    type Builder = TlsAcceptorBuilder;
 
     fn accept<S>(&self, stream: S)
             -> result::Result<tls_api::TlsStream<S>, tls_api::HandshakeError<S>>
